#include <stdio.h>
#include <string.h>

#include "plfunc.h"

Object reduce(Object* params, int length, struct Environment* env)
{
    Object list = params[0];
    const Object func = params[1];
    Object total = params[2];

    if (list.type != TYPE_LIST) {
        return simpleFuncEval(func, total, list, env);
    }

    FOR_POINTER_IN_LIST(&list) {
        total = simpleFuncEval(func, total, *POINTER, env);
    }

    return total;
}

Object charAt(Object* params, int length, struct Environment* env)
{
    Object string = params[0];
    Object at = params[1];

    char* c = malloc(sizeof(char) * 2);
    c[0] = string.string[at.number];
    c[1] = '\0';
    string.string = c;
    return string;
}

Object filter(Object* params, int length, struct Environment* env)
{
    Object condition = params[0];
    Object list = params[1];

    Object filtered = listObject();
    FOR_POINTER_IN_LIST(&list) {
        Object l = startList(cloneObject(condition));
        Object testee = cloneObject(*POINTER);
        nf_addToList(&l, testee);
        Object e = eval(&l, env);
        if (e.number) {
            nf_addToList(&filtered, testee); // May need to re-clone testee?
        }
        cleanObject(&l);
    }
    return filtered;
}

Object append(Object* params, int length, struct Environment* env)
{
    Object list = params[0];
    Object newElement = params[1];

    Object newList = cloneObject(list);
    nf_addToList(&newList, cloneObject(newElement));
    return newList;
}

Object prepend(Object* params, int length, struct Environment* env)
{
    Object list = params[0];
    Object newElement = params[1];

    Object newList = listObject();
    nf_addToList(&newList, cloneObject(newElement));
    appendList(&newList, &list);
    return newList;
}

Object at(Object* params, int length, struct Environment* env)
{
    Object index = params[0];
    Object list = params[1];

    const Object* found = itemAt(&list, index.number);
    if (found) {
        return cloneObject(*found);
    } else {
        return errorObject(INDEX_PAST_END);
    }
}

Object rest(Object* params, int length, struct Environment* env)
{
    Object list = params[0];

    Object ret = listObject();
    Object* l = &list;
    FOR_POINTER_IN_LIST(l) {
        if (POINTER == l->list) {
            continue;
        }
        nf_addToList(&ret, cloneObject(*POINTER));
    }
    return ret;
}

Object reverse(Object* params, int length, struct Environment* ignore2)
{
    Object _list = params[0];

    if (!isListy(_list)) {
        return errorObject(NOT_A_LIST);
    }
    const Object* list = &_list;
    Object rev = listObject();

    Object* tail = NULL;
    FOR_POINTER_IN_LIST(list) {
        Object* oldTail = tail;
        allocObject(&tail, cloneObject(*POINTER));
        if (oldTail) {
            tail->forward = oldTail;
        }
    }

    rev.list = tail;
    return rev;
}

Object isNum(Object* params, int length, struct Environment* env)
{
    Object test = params[0];

    return test.type == TYPE_NUMBER ? boolObject(1) : boolObject(0);
}

Object isList(Object* params, int length, struct Environment* env)
{
    Object test = params[0];

    return test.type == TYPE_LIST ? boolObject(1) : boolObject(0);
}

Object isString(Object* params, int length, struct Environment* env)
{
    Object test = params[0];

    return test.type == TYPE_STRING ? boolObject(1) : boolObject(0);
}

// Get the int value of a string's first character
Object charVal(Object* params, int length, struct Environment* env)
{
    Object test = params[0];

    return numberObject(test.string[0]);
    //    return test.type == TYPE_STRING && test.string[0] == '\0' ?
    //        boolObject(1) : boolObject(0);
}

Object isErr(Object* params, int length, struct Environment* env)
{
    Object test = params[0];

    return test.type == TYPE_ERROR ? boolObject(1) : boolObject(0);
}

Object parseEvalO(Object* params, int length, struct Environment* env)
{
    Object text = params[0];

    if (text.type == TYPE_SYMBOL) {
        Object string = eval(&text, env);
        Object parsed = parseEval(string.string, env);
        cleanObject(&string);
        return parsed;
    } else if (text.type == TYPE_SLIST) {
        return evalList(&text, env);
    } else if (text.type != TYPE_STRING) {
        return errorObject(CAN_ONLY_EVAL_STRINGS);
    }
    return parseEval(text.string, env);
}

Object listEquality(const Object* list1, const Object* list2)
{
    FOR_POINTERS_IN_LISTS(list1, list2) {
        if (P1->type != P2->type || P1->number != P2->number) {
            return boolObject(0);
        }
    }
    return boolObject(1);
}

#define CAT_MAX 1024

Object _catObjects(Object obj1, Object obj2, struct Environment* env)
{
    Object evalObj1 = eval(&obj1, env);
    Object evalObj2 = eval(&obj2, env);

    char str1[CAT_MAX] = "";
    char str2[CAT_MAX] = "";
    int length = stringObj(str1, &evalObj1);
    length += stringObj(str2, &evalObj2);

    cleanObject(&evalObj1);
    cleanObject(&evalObj2);

    Object o = newObject(TYPE_STRING);
    o.string = malloc(sizeof(char) * length + 1);
    sprintf(o.string, "%s%s", str1, str2);

    return o;
}

Object catObjects(Object* params, int length, struct Environment* env)
{
    Object string = stringFromSlice("", 0);
    if (length == 0) {
        return string;
    }
    for (int i = 0; i < length; i++) {
        Object newString = _catObjects(string, params[i], env);
        cleanObject(&string);
        string = newString;
    }
    return string;
}

Object _basicOp(const Object* obj1, const Object* obj2, const char op,
                struct Environment* env)
{
    const int n1 = obj1->number;
    const int n2 = obj2->number;

    switch (op) {
        case '+':
            if (eitherIsNot(TYPE_NUMBER, obj1, obj2)) {
                Object objects[2] = {*obj1, *obj2};
                return catObjects(objects, 2, env);
            }
            return numberObject(n1 + n2);
        case '-':
            return numberObject(n1 - n2);
        case '*':
            return numberObject(n1 * n2);
        case '/':
            return numberObject(n1 / n2);
        case '%':
            return numberObject(n1 % n2);

        case '&':
            return boolObject(n1 != 0 && n2 != 0);
        case '|':
            return boolObject(n1 != 0 || n2 != 0);

        case '=':
            if (bothAre(TYPE_STRING, obj1, obj2)) {
                return boolObject(!strcmp(obj1->string, obj2->string));
            }
            if (bothAre(TYPE_LIST, obj1, obj2)) {
                return listEquality(obj1, obj2);
            }
            return boolObject(n1 == n2 && areSameType(obj1, obj2));
        case '>':
            return boolObject(n1 > n2);
        case '<':
            return boolObject(n1 < n2);
        default:
            return *obj1;
    }
}

Object basicOp(const Object* obj1, const Object* obj2, const char op,
               struct Environment* env)
{
    if (isError(*obj2, ONLY_ONE_ARGUMENT)) {
        return *obj2;
    }

    int lists = (obj1->type == TYPE_LIST) + (obj2->type == TYPE_LIST);
    if (lists == 0) {
        return _basicOp(obj1, obj2, op, env);

    } else if (lists == 1) { // Single operand is applied to each element in list
        const Object* listObj = (obj1->type == TYPE_LIST) ? obj1 : obj2;
        const Object* singleObj = (obj1->type == TYPE_LIST) ? obj2 : obj1;

        Object newList = listObject();
        FOR_POINTER_IN_LIST(listObj) {
            Object adding = eval(POINTER, env);
            nf_addToList(&newList, _basicOp(&adding, singleObj, op, env));
        }
        return newList;

    } else { // 2 lists with the op applied to matching indices of both lists
        if (listLength(obj1) == listLength(obj2)) {
            Object newList = listObject();
            FOR_POINTERS_IN_LISTS(obj1, obj2) {
                const Object ev1 = eval(P1, env);
                const Object ev2 = eval(P2, env);
                nf_addToList(&newList, _basicOp(&ev1, &ev2, op, env));
            }
            return newList;
        } else {
            return errorObject(LISTS_NOT_SAME_SIZE);
        }
    }
}

Object len(Object* params, int length, struct Environment* env)
{
    Object obj1 = params[0];

    Object o = numberObject(listLength(&obj1));
    if (o.number < 0) {
        return errorObject(NOT_A_LIST);
    }
    return o;
}

#define BASIC_OP(_name, _char)                                      \
    Object _name(Object* params, int length, struct Environment* env) \
    { \
        Object obj1 = params[0]; \
        Object obj2 = params[1]; \
        return basicOp(&obj1, &obj2, _char, env); \
    }

#define BASIC_MATH(_name, _op) \
Object _name(Object* params, int length, struct Environment* env) \
{ \
    if (length == 0) { \
        return numberObject(0); \
    } \
    Object sum = numberObject(params[0].number); \
    for (int i = 1; i < length; i++) { \
        sum.number _op params[i].number; \
    } \
    return sum; \
}

BASIC_MATH(add, +=)

BASIC_MATH(sub, -=)

BASIC_MATH(mul, *=)

BASIC_MATH(dvi, /=)

BASIC_MATH(mod, %=)

BASIC_OP(equ, '=')

BASIC_OP(gth, '>')

BASIC_OP(lth, '<')

BASIC_OP(and, '&')

BASIC_OP(or, '|')

#undef BASIC_OP

#ifdef STANDALONE

Object print(Object* params, int length, struct Environment* env)
{
    Object p = params[0];
    Object c = cloneObject(p);
    Object e = eval(&c, env);
    _printObj(&e, 0);
    cleanObject(&c);
    cleanObject(&e);
    return numberObject(0);
}

Object pChar(Object* params, int length, struct Environment* env)
{
    Object c = params[0];

    if (c.type != TYPE_NUMBER) {
        return errorObject(BAD_NUMBER);
    }
    printf("%c", c.number % 256);
    return numberObject(0);
}

Object printEnvO(Object* params, int length, struct Environment* env)
{
    printEnv(global());
    return numberObject(0);
}

Object takeInput(Object* params, int length, struct Environment* env)
{
    Object prompt = params[0];
    if (prompt.type == TYPE_STRING) {
        printf("%s", prompt.string);
    }
    char input[256] = "";
    fgets(input, 256, stdin);
    return stringFromSlice(input, strlen(input) - 1);
}

Object loadFile(Object* params, int length, struct Environment* env)
{
    Object filename = params[0];

    if (isStringy(filename)) {
        readFile(filename.string, env);
        return numberObject(0);
    }
    return numberObject(1);
}

Object systemCall(Object* params, int length, struct Environment* env)
{
    Object process = params[0];

    if (isStringy(process)) {
        return numberObject(system(process.string));
    }
    return numberObject(255);
}

Object help(Object* params, int length, struct Environment* env)
{
    Object symbol = params[0];

    char* help = getHelp(symbol.string);
    Object helpText = newObject(TYPE_STRING);
    helpText.string = help;
    return helpText;
}

char* readFileToString(FILE* input)
{
    size_t capacity = 128;
    char* string = malloc(sizeof(char) * capacity);
    int c;
    int i = 0;

    while ((c = fgetc(input)) != EOF) {
        string[i] = c;
        i++;
        if (i == capacity) {
            char* prev = string;
            capacity *= 2;
            string = malloc(sizeof(char) * capacity);
            memcpy(string, prev, sizeof(char) * capacity / 2);
            free(prev);
        }
    }

    return string;
}

Object readFileToObject(Object* params, int length, struct Environment* env)
{
    Object filename = params[0];

    if (filename.type != TYPE_STRING) {
        return errorObject(NULL_PARSE);
    }

    FILE* file = fopen(filename.string, "r");
    if (!file) {
        return errorObject(NULL_PARSE);
    }

    Object string = newObject(TYPE_STRING);
    string.string = readFileToString(file);
    return string;
}

#endif // STANDALONE
